// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import { TSignGetClient } from './tsignClient.js';
import TSignClient from './tsignClient.js';
import BigInt from 'big-integer';
import bitcoin from './bitcoin.js';

const p = BigInt('0fffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141', 16);

export default class TSignFlow {
    constructor(urls, user, isPost = false) {
        this.clients = urls.map(url => new (isPost ? TSignClient : TSignGetClient)(url, user));
        this.user = user;
        this.onProgress = () => {};
    }

    async sign(message) {
        try {
            var m = bitcoin.getM(message);
            
            this.onProgress(0, "starting");
            var ks = await Promise.all(this.clients.map((cln) => cln.getKi(m)));
            var alphaArgs = Array.from(ks.keys()).map(j => ks.map((ki, i) => ({ i, ki })).filter((_, i) => i !== j));
    
            this.onProgress(1, "processing");
            var alphas = await Promise.all(this.clients.map((cln, i) => cln.getAlpha(alphaArgs[i].map(arg => arg.ki))));
            var allAlphas = alphas.map((resp, i) => resp.public.map((alpha, j) => ({ j: alphaArgs[i][j].i, alpha }))) //zip alphas and is
                .reduce((alphas, alpha) => alphas.concat(alpha));
    
            this.onProgress(2, "processing");
            var deltas = await Promise.all(this.clients.map((cln, i) => cln.getDelta(alphas[i].private, allAlphas.filter(({ j, _ }) => i == j).map(({ _, alpha }) => alpha))));
            var allDeltas = deltas.map(resp => resp.public);
            
            this.onProgress(3, "processing");
            var signs = await Promise.all(this.clients.map((cln, i) => cln.getRSi(deltas[i].private, allDeltas)));
            var s = signs.reduce((sum, sign) => sum.add(sign.si).mod(p), BigInt.zero);
    
            var address = bitcoin.getAddressFromShares(signs);
            var signature = bitcoin.formatSign(m, signs[0].r, s, address);
    
            this.onProgress(4, "finished");
            return { address, signature };    
        } catch (err) {
            this.onProgress(4, "error");
            return Promise.reject(err);
        }
    }
}
